#ifndef __FRAGMENTEDMESSAGE_H__
#define __FRAGMENTEDMESSAGE_H__

#include<Arduino.h>

const int MESSAGE_FRAGMENT_SIZE = 30;
const int MESSAGE_FRAGMENT_RESERVED = 2;

const int FRAGMENT_SEQUENCING_INDEX = 0;
const int FRAGMENT_CHECKSUM_INDEX = 1;

//static PROGMEM prog_uint32_t b32_crc_table[16] = {
//    0x00000000, 0x1db71064, 0x3b6e20c8, 0x26d930ac,
//    0x76dc4190, 0x6b6b51f4, 0x4db26158, 0x5005713c,
//    0xedb88320, 0xf00f9344, 0xd6d6a3e8, 0xcb61b38c,
//    0x9b64c2b0, 0x86d3d2d4, 0xa00ae278, 0xbdbdf21c
//};

//http://hacromatic.com/blog/2012/08/make-your-serial-protocol-robust-with-8-bit-crc-codes/
static PROGMEM prog_uint8_t b8_crc_table[256] = {
  0x00, 0x2f, 0x5e, 0x71, 0xbc, 0x93, 0xe2, 0xcd, 0x57, 0x78, 0x09, 0x26, 0xeb, 0xc4, 0xb5, 0x9a,
  0xae, 0x81, 0xf0, 0xdf, 0x12, 0x3d, 0x4c, 0x63, 0xf9, 0xd6, 0xa7, 0x88, 0x45, 0x6a, 0x1b, 0x34,
  0x73, 0x5c, 0x2d, 0x02, 0xcf, 0xe0, 0x91, 0xbe, 0x24, 0x0b, 0x7a, 0x55, 0x98, 0xb7, 0xc6, 0xe9,
  0xdd, 0xf2, 0x83, 0xac, 0x61, 0x4e, 0x3f, 0x10, 0x8a, 0xa5, 0xd4, 0xfb, 0x36, 0x19, 0x68, 0x47,
  0xe6, 0xc9, 0xb8, 0x97, 0x5a, 0x75, 0x04, 0x2b, 0xb1, 0x9e, 0xef, 0xc0, 0x0d, 0x22, 0x53, 0x7c,
  0x48, 0x67, 0x16, 0x39, 0xf4, 0xdb, 0xaa, 0x85, 0x1f, 0x30, 0x41, 0x6e, 0xa3, 0x8c, 0xfd, 0xd2,
  0x95, 0xba, 0xcb, 0xe4, 0x29, 0x06, 0x77, 0x58, 0xc2, 0xed, 0x9c, 0xb3, 0x7e, 0x51, 0x20, 0x0f,
  0x3b, 0x14, 0x65, 0x4a, 0x87, 0xa8, 0xd9, 0xf6, 0x6c, 0x43, 0x32, 0x1d, 0xd0, 0xff, 0x8e, 0xa1,
  0xe3, 0xcc, 0xbd, 0x92, 0x5f, 0x70, 0x01, 0x2e, 0xb4, 0x9b, 0xea, 0xc5, 0x08, 0x27, 0x56, 0x79,
  0x4d, 0x62, 0x13, 0x3c, 0xf1, 0xde, 0xaf, 0x80, 0x1a, 0x35, 0x44, 0x6b, 0xa6, 0x89, 0xf8, 0xd7,
  0x90, 0xbf, 0xce, 0xe1, 0x2c, 0x03, 0x72, 0x5d, 0xc7, 0xe8, 0x99, 0xb6, 0x7b, 0x54, 0x25, 0x0a,
  0x3e, 0x11, 0x60, 0x4f, 0x82, 0xad, 0xdc, 0xf3, 0x69, 0x46, 0x37, 0x18, 0xd5, 0xfa, 0x8b, 0xa4,
  0x05, 0x2a, 0x5b, 0x74, 0xb9, 0x96, 0xe7, 0xc8, 0x52, 0x7d, 0x0c, 0x23, 0xee, 0xc1, 0xb0, 0x9f,
  0xab, 0x84, 0xf5, 0xda, 0x17, 0x38, 0x49, 0x66, 0xfc, 0xd3, 0xa2, 0x8d, 0x40, 0x6f, 0x1e, 0x31,
  0x76, 0x59, 0x28, 0x07, 0xca, 0xe5, 0x94, 0xbb, 0x21, 0x0e, 0x7f, 0x50, 0x9d, 0xb2, 0xc3, 0xec,
  0xd8, 0xf7, 0x86, 0xa9, 0x64, 0x4b, 0x3a, 0x15, 0x8f, 0xa0, 0xd1, 0xfe, 0x33, 0x1c, 0x6d, 0x42
};

class FragmentedMessage
{
public:
  int NumFragments;
  char** Fragments;
  char* ChecksumFragment;

  FragmentedMessage()
  {
    NumFragments = 15;
    Fragments = new char*[NumFragments];
    for(int i =0; i < NumFragments; i++)
    {
      Fragments[i] = NULL;
    }

    ChecksumFragment = NULL;
  }

  FragmentedMessage(char* message, int senderID)
  {
    int numFragments = (strlen(message) / MESSAGE_FRAGMENT_SIZE) + 1;
    char** fragments = new char*[numFragments];
    memset(fragments, 0, numFragments);

    for(int fragmentIndex = 0; fragmentIndex < numFragments; fragmentIndex++)
    {
      char* fragment = new char[MESSAGE_FRAGMENT_SIZE + MESSAGE_FRAGMENT_RESERVED];
      memset(fragment, 0, MESSAGE_FRAGMENT_SIZE + MESSAGE_FRAGMENT_RESERVED);


      strncpy((fragment + MESSAGE_FRAGMENT_RESERVED), (message + (MESSAGE_FRAGMENT_SIZE * fragmentIndex)), MESSAGE_FRAGMENT_SIZE);

      fragment[FRAGMENT_SEQUENCING_INDEX] = (char)((fragmentIndex << 4) | (senderID % 16));
      fragment[FRAGMENT_CHECKSUM_INDEX] = crc_8bit(fragment + MESSAGE_FRAGMENT_RESERVED, MESSAGE_FRAGMENT_SIZE);

      fragments[fragmentIndex] = fragment;
      //debugPrintMessageFragment(fragment);
    }

    char* checksumFragment = createChecksumFragment(message, numFragments, senderID);

    //DebugPrintMessageFragment(checksumFragment);

    Fragments = fragments;
    NumFragments = numFragments;
    ChecksumFragment = checksumFragment;
  }

  ~FragmentedMessage()
  {
    for(int i = 0; i < NumFragments; i++)
    {
      if(Fragments[i] != NULL)
      {
        delete[] Fragments[i]; 
      }
    }

    if(ChecksumFragment != NULL)
    {
      delete[] ChecksumFragment;
    }
    
    if(Fragments != NULL)
    {
      delete[] Fragments;
    }
  }
  
  void DebugPrintMessageFragment(char* messageFragment)
  {
    //return;
    
     int senderID = DecodeFragmentSenderID(messageFragment);
     int fragmentNumber = DecodeFragmentNumber(messageFragment);
     int fragmentChecksum = DecodeFragmentChecksum(messageFragment);
     
     Serial.print("Sender: ");
     Serial.print(senderID);
     Serial.print(" Fragment: ");
     Serial.print(fragmentNumber);
     Serial.print(" Checksum: ");
     Serial.println(fragmentChecksum);
     

    Serial.print("[");

    for(int i = MESSAGE_FRAGMENT_RESERVED; i < MESSAGE_FRAGMENT_SIZE + MESSAGE_FRAGMENT_RESERVED; i++)
    {
      Serial.print((char)messageFragment[i]);
    }

    Serial.println("]");
    
    Serial.println("");
  }
  
  char* DefragmentMessage()
  {
    int activeFragments = countActiveFragments();
    int expectedFragments = getExpectedFragmentsFromFragmentedMessage();

    //Serial.print("Expected: ");
    //Serial.print(expectedFragments);
    //Serial.print(" Active: ");
    //Serial.println(activeFragments);

    if(activeFragments != expectedFragments)
    {
       return NULL;
    }

    char* messageBuffer = new char[MESSAGE_FRAGMENT_SIZE * activeFragments];
    memset(messageBuffer, 0, MESSAGE_FRAGMENT_SIZE * activeFragments);

    for(int fragmentIndex = 0; fragmentIndex < activeFragments; fragmentIndex++)
    {
      if(Fragments[fragmentIndex] == NULL)
      {
        continue;
      }

      //DebugPrintMessageFragment(Fragments[fragmentIndex]);

      strncpy((messageBuffer + (MESSAGE_FRAGMENT_SIZE * fragmentIndex)), (Fragments[fragmentIndex] + MESSAGE_FRAGMENT_RESERVED), MESSAGE_FRAGMENT_SIZE);
      //Serial.println(messageBuffer);
    }

    //Serial.print("Defragmented: {");
    //Serial.print(messageBuffer);
    //Serial.println("}");
    
    char bufferChecksum = crc_8bit(messageBuffer, MESSAGE_FRAGMENT_SIZE * activeFragments);
    int recvdMessageChecksum = getChecksumFromFragmentedMessage();

    //Serial.print("buffer chk: ");
    //Serial.print(bufferChecksum);
    //Serial.print(" recvd chk: ");
    //Serial.println(recvdMessageChecksum);
      
    if(bufferChecksum != recvdMessageChecksum)
    {
      delete[] messageBuffer;
      return NULL;
    }
    
    return messageBuffer;
  }
  
  int HasChecksumFragment()
  {
    return ChecksumFragment != NULL;
  }
  
  void AppendMessageFragment(char* messageFragment)
  {
    int recvdFragmentChecksum = DecodeFragmentChecksum(messageFragment);
    char actualFragmentChecksum = crc_8bit(messageFragment + MESSAGE_FRAGMENT_RESERVED, MESSAGE_FRAGMENT_SIZE);

    if(recvdFragmentChecksum != actualFragmentChecksum)
    {
      return;
    }
    

    char* fragmentCopy = new char[MESSAGE_FRAGMENT_SIZE + MESSAGE_FRAGMENT_RESERVED];
    memcpy(fragmentCopy, messageFragment, MESSAGE_FRAGMENT_SIZE + MESSAGE_FRAGMENT_RESERVED);

    int fragmentNumber = DecodeFragmentNumber(fragmentCopy);

    if(fragmentNumber == 0x0F)
    {
      if(ChecksumFragment != NULL)
      {
        delete[] ChecksumFragment;
      }
      ChecksumFragment = fragmentCopy; 
    }
    else    
    {
      if(Fragments[fragmentNumber] != NULL)
      {
        delete[] Fragments[fragmentNumber];
      }
      Fragments[fragmentNumber] = fragmentCopy; 
    }
  }
  
  static int DecodeFragmentSenderID(char* message)
  {
    return 0x0F & message[FRAGMENT_SEQUENCING_INDEX];
  }
  
  static int DecodeFragmentNumber(char* message)
  {
    return (message[FRAGMENT_SEQUENCING_INDEX] >> 4) & 0x0F;
  }

  static int DecodeFragmentChecksum(char* message)
  {
    return message[FRAGMENT_CHECKSUM_INDEX];
  }

private:
  char* createChecksumFragment(char* message, int numFragments, int senderID)
  {
    char* fragment = new char[MESSAGE_FRAGMENT_SIZE + MESSAGE_FRAGMENT_RESERVED];
    memset(fragment, 0, MESSAGE_FRAGMENT_SIZE + MESSAGE_FRAGMENT_RESERVED);

    char checksum = crc_8bit(message, MESSAGE_FRAGMENT_SIZE * numFragments);

    snprintf(fragment + MESSAGE_FRAGMENT_RESERVED, MESSAGE_FRAGMENT_SIZE, "%d|%d", checksum, numFragments);

    fragment[FRAGMENT_SEQUENCING_INDEX] = (char)(0xF0 | (senderID % 16));
    fragment[FRAGMENT_CHECKSUM_INDEX] = crc_8bit(fragment + MESSAGE_FRAGMENT_RESERVED, MESSAGE_FRAGMENT_SIZE);

    return fragment;
  }

//  char calculateChecksum(char* message, int byteCount)
//  {
//    char checksum = 0;
//    for(int i = 0; i < byteCount; i++)
//    {
//      checksum ^= message[i];
//    }
//
//    return checksum;
//  }
  
  int getChecksumFromFragmentedMessage()
  {
    int checksum;
    sscanf(ChecksumFragment + MESSAGE_FRAGMENT_RESERVED, "%d|%*d", &checksum);
    return checksum;
  }

  int getExpectedFragmentsFromFragmentedMessage()
  {
    int expectedFragments;
    sscanf(ChecksumFragment + MESSAGE_FRAGMENT_RESERVED, "%*d|%d", &expectedFragments);
    return expectedFragments;
  }
  
  int countActiveFragments()
  {
    int activeFragments = 0;
    for(; (activeFragments < NumFragments) && (Fragments[activeFragments] != NULL); activeFragments ++) {}
    return activeFragments;
  }

char crc_8bit(char *data, size_t data_len)
{
    char crc = 0;

    unsigned int tbl_idx;

    while (data_len--) {
        tbl_idx = ((crc >> 0) ^ *data) & 0xff;
        crc = (pgm_read_byte(b8_crc_table+tbl_idx) ^ (crc << 8)) & 0xff;

        data++;
    }
    return crc & 0xff;
}

//http://excamera.com/sphinx/article-crc.html


//unsigned long b32_crc_update(unsigned long crc, byte data)
//{
//    byte tbl_idx;
//    tbl_idx = crc ^ (data >> (0 * 4));
//    crc = pgm_read_dword_near(b32_crc_table + (tbl_idx & 0x0f)) ^ (crc >> 4);
//    tbl_idx = crc ^ (data >> (1 * 4));
//    crc = pgm_read_dword_near(b32_crc_table + (tbl_idx & 0x0f)) ^ (crc >> 4);
//    return crc;
//}

//unsigned long crc_32bit(char *s, size_t data_len)
//{
//  unsigned long crc = ~0L;
//  while (data_len--)
//  {
//    crc = b32_crc_update(crc, *s++);
//  }
//  crc = ~crc;
//  return crc;
//}



};

#endif

